import numpy as np
from numba import njit
#import cProfile  
from numba import complex128, int32
from qshap.utils import inv_binom_coef, complex_v_invc_degree, store_complex_v_invc, complex_dot_v2, weight, store_complex_root


@njit
def T2_sample(i, w_matrix, w_ind, init_prediction, store_v_invc, store_z, shap_value, feature_uniq):
    ## Calculate T2 for each sample
    L, p = w_matrix.shape
     
    for l1 in range(L):
        for l2 in range(l1, L):
            init_prediction_product = init_prediction[l1] * init_prediction[l2]
            
            union_f12 = feature_uniq[(w_ind[l1, feature_uniq] + w_ind[l2, feature_uniq]) >= 1]
            
            n12 = len(union_f12)
            # begin to use the property of complex conjugate
            n12_c = n12 // 2 + 1

            v_invc = store_v_invc[n12, :n12_c]
            z = store_z[n12, :n12_c]

            p_z = np.zeros((n12_c), dtype=complex128)
            tmp_p_z = np.zeros((n12_c), dtype=complex128)

            for k in range(n12_c):
                p_z[k] = np.prod(z[k] + w_matrix[l1, union_f12] * w_matrix[l2, union_f12])

            # update only when feature j belongs to the union of f1 and f2
            for j in union_f12:
                    # remove the operation for j by dividing
                for k in range(n12_c):
                    tmp_p_z[k] = p_z[k] / (z[k] + w_matrix[l1, j] * w_matrix[l2, j])

                if l1 != l2:
                    shap_value[i, j] += 2 * (w_matrix[l1, j] * w_matrix[l2, j] - 1) * init_prediction_product * complex_dot_v2(tmp_p_z, v_invc, n12)
                else:
                    shap_value[i, j] += (w_matrix[l1, j] * w_matrix[l2, j] - 1) * init_prediction_product * complex_dot_v2(tmp_p_z, v_invc, n12)

                    
def T2(x, summary_tree, store_v_invc, store_z, parallel = True):
    """
    Calculate the second order treeshap value
    
    Parameters:
    -x: sample to be explained
    -summary_tree: summary tree
    
    Return:
    treeshap value for the sample
    """
    
    init_prediction = summary_tree.init_prediction[summary_tree.init_prediction!=0]
    
    shap_value = np.zeros_like(x)
    
    for i in range(x.shape[0]):
        xi = x[i]
        w = weight(xi, summary_tree)
        w_matrix = w[0]
        w_ind = w[1]

        T2_sample(i, w_matrix, w_ind, init_prediction, store_v_invc, store_z, shap_value, summary_tree.feature_uniq)

    return shap_value


def loss_treeshap(x, y, summary_tree, store_v_invc, store_z, explainer, learning_rate=1):
    """
    Explain l2 loss for every sample
    
    Parameters:
    -x: samples
    -y: y corresponding to x
    -summary_tree: summary tree
    -store_v_invc: stored v_invc/d
    -store_z: stored complex root of unity 
    -explainer: explainer from shap
    -learning_rate: learning_rate if it's a tree from scikit learn GBM, learning_rate for decision tree and xgboost should be 1. 
    
    Return:
    loss treeshap for x
    """

    square_treeshap_x = T2(x, summary_tree, store_v_invc, store_z) * learning_rate ** 2 
    # direct call from shap
    T0_x = explainer.shap_values(x) * learning_rate 
    res = square_treeshap_x - 2 * (y * T0_x.T).T
    return res

# WARNING: This is the wrong way to parallelize the code! The overhead is even more severe for ensemble trees, and it only serves as a bad example to warn the successors to NOT parallelize in this way.
# def loss_treeshap_parallel(x, y, summary_tree, store_v_invc, store_z, explainer, learning_rate=1, ncore=-1):
#     """
#     Explain l2 loss for every sample
    
#         Explain l2 loss for every sample
    
#     Parameters:
#     -x: samples
#     -y: y corresponding to x
#     -summary_tree: summary tree
#     -store_v_invc: stored v_invc/d
#     -store_z: stored complex root of unity 
#     -explainer: explainer from shap
#     -learning_rate: learning_rate if it's a tree from scikit learn GBM, learning_rate for decision tree and xgboost should be 1. 
#     -ncore: number of cores to use, with default value -1 to utilize all the cores
    
#     Return:
#     loss treeshap for x
#     """
#     max_core = os.cpu_count()
#     if ncore == -1:
#         ncore = os.cpu_count()
#     ncore = min(max_core, ncore)
    
#     if ncore==1:
#         square_treeshap_x = T2(x, summary_tree, store_v_invc, store_z) * learning_rate ** 2 
#     else:
#         chunks = divide_chunks(x, ncore)

#         # Use ProcessPoolExecutor to process each chunk in parallel
#         with ProcessPoolExecutor(max_workers=ncore) as executor:
#             # Submit all chunks for processing
#             futures = [executor.submit(T2, chunk, summary_tree, store_v_invc, store_z) for chunk in chunks]

#             # Wait for all futures to complete and collect results
#             results = [future.result() for future in futures]

#         # Combine the results back into a single array
#         # have to compensate the learning_rate
#         square_treeshap_x = np.concatenate(results) * learning_rate ** 2      
        
#     # direct call from shap
#     T0_x = explainer.shap_values(x) * learning_rate 
#     res = square_treeshap_x - 2 * (y * T0_x.T).T
#     return res




import numpy as np
from numba import njit
import time as time
from dataclasses import dataclass
import json
#d_test = 9

def inv_binom_coef(d):
    """
    calculate the inverse of binomial coefficients using an iterative method with symmetry.
    
    Parameters:
    - d: dimension
    
    Example: 
    binom_coef(5)
    ([ 1.,  5., 10., 10.,  5.,  1.])
    """
    coef = np.zeros(d + 1)
    coef[0] = 1
    for i in range(1, d // 2 + 1):
        coef[i] = coef[i - 1] * (d - i + 1) / i
    for i in range(d // 2 + 1, d + 1):
        coef[i] = coef[d - i]
    return 1/coef

#inv_binom_coef(d_test)

def complex_v_invc_degree(d):
    """
    Pre store v_invc: v(z)^-1 @ c / d at degree d where z are complex roots of unity
    
    Parameters:
    -d: degree
    -c: the coefficients matrix
    """
    omega_inv = np.exp(-2 * np.pi * 1j * np.arange(d) / d)
    v_omega_inv = np.vander(omega_inv, increasing=True)
    v_inv_omega_theo = v_omega_inv / d
    res = v_inv_omega_theo @ inv_binom_coef(d-1) / d
    return res

#print(complex_v_invc_degree(d_test))

def store_complex_v_invc(d):
    """
    Pre store v_invc: v(z)^-1 @ c / d up to maximum tree depth where z are complex roots of unity
    
    Parameters:
    -d: max treedepth
    """
    res = np.zeros((d+1, d), dtype=complex)
    
    for i in range(1, d+1):
        res[i, :i] = complex_v_invc_degree(i)
    
    return res 

# how to retrive the degree d_test pre_stored value
# print(store_complex_v_invc(20)[d_test, :d_test])

def store_complex_root(d):
    """
    Prestore the complex root of unity z
    
    Parameters:
    -d: max treedepth
    """ 
    res = np.zeros((d+1, d), dtype=complex)
    
    for i in range(1, d+1):
        res[i, :i] = np.exp(2 * np.pi * 1j * np.arange(i) / i)
    return res


@njit
def complex_dot_v2(p, v_invc, d):
    """
    Return the dot product: C(z) * P(z) / d where z are the complex roots of unity, using the fact that P(w) and v_invc both:
    except for the 0 index and possibly the last when d is odd, the rest are complex conjugate by head and tail, etc...
    
    Parameters 
    - p: a polynomial vector evaluated at complex root of unity
    - v_invc: pre-calculated inverse coefficients
    - d: the original degree before cut by half.
    """
    len_p = len(p)
    res = p[0] * v_invc[0]
    if d % 2 == 0:
        res += 2 * np.dot(p[1:(len_p-1)], v_invc[1:(len_p-1)]) + p[-1] * v_invc[-1]
    else:
        res += 2 * np.dot(p[1:len_p], v_invc[1:len_p])
    return res.real

# v1 = np.array([3, 6-1j, 5+1j, 6+1j])
# v2 = np.array([5, 4-1j, 6+1j, 4+1j])
# v3 = np.array([3, 6-1j, 5+1j])
# v4 = np.array([5, 4-1j, 6+1j])
# print(complex_dot_v2(v3, v4, len(v1)))

# v5 = np.array([3, 6-1j, 5+1j, 5-1j, 6+1j])
# v6 = np.array([5, 4-1j, 6+1j, 6-1j, 4+1j])
# v7 = np.array([3, 6-1j, 5+1j])
# v8 = np.array([5, 4-1j, 6+1j])
# print(complex_dot_v2(v7, v8, len(v5)))


@dataclass(frozen=True)
class simple_tree:
    """
    dataclass for a simple tree, in the scikit learn format that is necessary for computation
    
    Data:
    children_left: left children_index
    children_right: right children_index
    feature: array of features splitted at each node
    threshold: array of threshols for corresponding splitting features
    max_depth: max_depth of the tree
    n_node_samples: array of sample size for each node
    value: array of values for each node, only leaf value is used, so only keep leaf value is fine
    node_count: total number of leaves 
    """
    children_left: np.ndarray
    children_right: np.ndarray
    feature: np.ndarray
    threshold: np.ndarray
    max_depth: int
    n_node_samples: np.ndarray
    value: np.ndarray
    node_count: int


@dataclass(frozen=True)
class tree_summary:
    """
    dataclass for the calculation of cd-treeshap family
    
    Data:
    - children_left: left children_index
    - children_right: right children_index 
    - feature: array of features splitted at each node
    - feature_uniq: array of uniq features
    - threshold: array of threshols for corresponding splitting features
    - max_depth: the max_depth of the tree
    - sample_weight: list of sample size of parent/sample size of current node
    - init_prediction: initial prediction from each leaf
    - node_count: number of nodes
    """
    children_left: np.ndarray
    children_right: np.ndarray
    feature: np.ndarray
    feature_uniq: np.ndarray
    threshold: np.ndarray
    max_depth: int
    sample_weight: np.ndarray
    init_prediction: np.ndarray
    node_count: int
    

    
def summarize_tree(tree):
    """
    Summarize the data needed for tree_summary. The tree object should have:
    children_left: left children_index
    children_right: right children_index
    feature: array of features splitted at each node
    threshold: array of threshols for corresponding splitting features
    max_depth: max_depth of the tree
    n_node_samples: array of sample size for each node
    value: array of values for each node, only leaf value is used, so only keep leaf value is fine
    node_count: total number of leaves 
    """
    sample_weight = np.ones_like(tree.threshold)
    init_prediction = np.zeros_like(tree.threshold)
    n = tree.n_node_samples[0]
    
    def traversal_summarize_tree(v):
        v_l, v_r = tree.children_left[v], tree.children_right[v]
        n_v, n_l, n_r = tree.n_node_samples[v], tree.n_node_samples[v_l], tree.n_node_samples[v_r]
        
        if v_l < 0:  #leaf
            init_prediction[v] = tree.value[v] * n_v/n
        else:
            sample_weight[v_l], sample_weight[v_r] = n_v/n_l, n_v/n_r
            traversal_summarize_tree(v_l)
            traversal_summarize_tree(v_r)
    
    # travel from the root
    traversal_summarize_tree(0)
    
    feature_uniq = np.unique(tree.feature[tree.feature >= 0])
   
    return tree_summary(tree.children_left, tree.children_right, tree.feature, feature_uniq, tree.threshold, tree.max_depth, sample_weight, init_prediction, tree.node_count)


@njit
def traversal_weight(x, v, w, children_left, children_right, feature, threshold, sample_weight, leaf_ind, w_res, w_ind, depth, met_feature):
    """
    Calculate the weight in the treeSHAP. 

    Parameters:
    - x: one sample to be explained 
    - v: node index 
    - w: weight vector passed to the current node, for temporary usage
    - children_left: left children_index
    - children_right: right children_index 
    - feature: list of features splitted at each node
    - threshold: list of threshold for corresponding features
    - sample_weight: a list of sample size of parent/sample size of current node
    - leaf_ind: leaf indices
    - w_res, L * p matrix of weights, which records the modified weight for each leaf and each feature
    - w_ind, L * p matrix of indicator matrix, which records the met of features for each leaf
    - depth: current depth
    - met_feature: record all the features met to now

    Update:
    w_res
    w_ind
    met_feature
    """

    v_l, v_r = children_left[v], children_right[v]

    if v_l < 0:
        # match to the right location so the value for w_res corresponds to the same order of leaf_ind
        ind = (leaf_ind == v)
        #feature_tmp = met_feature[0:depth]
        for tmp_depth in range(depth):
            w_res[ind, met_feature[tmp_depth]] = w[tmp_depth]
            w_ind[ind, met_feature[tmp_depth]] = 1
    else:
        split_feature = feature[v]
        split_threshold = threshold[v]

        former_depth = np.arange(depth)[met_feature[0:depth] == split_feature]

        if len(former_depth) != 0:
            former_depth = former_depth[-1]
        else:
            former_depth = depth  
            w[depth] = 1

        met_feature = met_feature.copy()
        met_feature[depth] = split_feature

        w_r = w.copy()

        if x[split_feature] <= split_threshold:
            w[depth] = w[former_depth] * sample_weight[v_l]
            w_r[depth] = 0
        else:
            w_r[depth] = w_r[former_depth] * sample_weight[v_r]
            w[depth] = 0 

        traversal_weight(x, v_l, w, children_left, children_right, feature, threshold, sample_weight, leaf_ind, w_res, w_ind, depth+1, met_feature) 
        traversal_weight(x, v_r, w_r, children_left, children_right, feature, threshold, sample_weight, leaf_ind, w_res, w_ind, depth+1, met_feature)

        
def weight(x, summary_tree):
    p = len(x)
    d = summary_tree.max_depth
    
    feature_uniq = summary_tree.feature_uniq
    
    leaf_ind = np.arange(summary_tree.node_count)[summary_tree.children_left==-1]

    # L * p matrix. Note that unused features are also labeled as 0 here for efficient storage.
    w_res = np.empty((len(leaf_ind), p))
    w_res[:, feature_uniq] = 1
    
    w = np.empty(d)
    
    # L * p matrix. [i, j] = 1 if the feature j is used by leaf i. [i, j] = 0 corresponds to 1 in the
    # above matrix. This two sparse matrices together could make the weight matrix well-defined, and save the
    # storage at the same time
    w_ind = np.empty((len(leaf_ind), p))
    w_ind[:, feature_uniq] = 0
    
    met_feature = np.full(d, -1, dtype=int)
    
    # begin traversal from root
    traversal_weight(x, 0, w, summary_tree.children_left, summary_tree.children_right, summary_tree.feature, summary_tree.threshold, summary_tree.sample_weight, leaf_ind, w_res, w_ind, 0, met_feature)
    return w_res, w_ind


def xgb_formatter(model_data, max_depth):
    """
    This function takes the json format of the xgboost output and transform it to a list that treeshap rsq can understand
    
    Parameters:
    model_data: json file
    max_depth: the max tree depth
    
    Examples:
    import json
    xgb_regressor.save_model("model.json")
    with open('model.json', 'r') as file:
    model_data = json.load(file)
    xgb_tree_res = xgb_formatter(model_data, 4)
    """
    trees_data = model_data["learner"]["gradient_booster"]["model"]["trees"]

    xgb_tree = []

    for tree in trees_data:
        xgb_tree.append(simple_tree(np.array(tree["left_children"]),
                                    np.array(tree["right_children"]), 
                                    np.array(tree["split_indices"]),
                                    np.array(tree["split_conditions"]),
                                    max_depth, 
                                    np.array(tree["sum_hessian"]), 
                                    np.array(tree["base_weights"]), 
                                    int(tree["tree_param"]["num_nodes"])))

    return(xgb_tree)



# Define a function to divide the dataset into chunks
def divide_chunks(data, n_chunks):
    total_elements = data.shape[0]
    chunk_size = total_elements // n_chunks
    remainder = total_elements % n_chunks

    chunks = []
    for i in range(n_chunks):
        start_index = i * chunk_size
        # For the last chunk, add the remainder to it
        end_index = start_index + chunk_size + (remainder if i == n_chunks - 1 else 0)
        chunks.append(data[start_index:end_index])

    return chunks




import ipywidgets as widgets
import numpy as np
import matplotlib.pyplot as plt
from ipywidgets import BoundedIntText, Layout, interactive_output, VBox, HBox
from IPython.display import display, clear_output

class vis:

    @staticmethod
    def rsq(x, color_map_name="Blues", horizontal=False, model_rsq=True, max_feature=10, cutoff=0, title="Shapley R²", xtitle="Feature index", ytitle="R²", rotation=0, label=None, decimal=3, save_name=None):
        """
        Visualize shapley rsq
        
        Parameters:
        -x: 1 dim array, but we recommend array from shapley rsq OvO ...
        -color_map_name: Color map allows to you take a variety color map from matplotlib to customize your visualization color.
        You can try to use "Pastel1", "Pasterl2", "PuBu", "Reds", "Greens" ...
        -horizontal: horizontal plot or not
        -model_rsq: display model rsq or not
        -max_feature: maximum number of features to show
        -cutoff: only value greater than or equal to the cutoff will be displayed
        -xtitle: xtitle, note that it's reversed for horizontal 
        -ytitle: ytitle, note that it's reversed for horizontal 
        -rotation: rotation of the tick
        -label: label for the features
        -title: plot title
        -decimal: decimals to show 
        -save_name: the name of the file if you want to save. None as default without saving.
        """
        
        # Calculate the sum of x
        x_sum = np.sum(x)
        x_len = np.size(x)
        cutoff_feature = np.sum(x >= cutoff)
        show_len = min(x_len, max_feature, cutoff_feature)

        # Sort x and keep track of the indices
        indices = np.argsort(x)[::-1]  # Indices of sorted elements (in descending order)
            
        sorted_x = x[indices]          # Sorted x
        indices = indices[:show_len]
        if label is not None:
            if len(x)!=len(label):
            #error: the length of label and and x must match 
                raise ValueError("The length of the label and x must match")
            else:
                sorted_label = label[indices]
                sorted_label = sorted_label[:show_len]
            
        sorted_x = sorted_x[:show_len]

        # Creating a color map based on the sorted values
        # Normalize the sorted_x values to the range [0, 1]
        normalized_x = (sorted_x - min(sorted_x)) / (max(sorted_x) - min(sorted_x))

        # Get the colormap
        cmap = plt.get_cmap(color_map_name)

        # Generate colors from the colormap
        colors = cmap(normalized_x)
        
        # Define an offset for the text to ensure it doesn't touch the bars
        text_offset = max(sorted_x) * 0.02  # 2% of the max value as the offset
        
        if not horizontal:
            # Create the bar chart
            bars = plt.bar(range(len(sorted_x)), sorted_x, color=colors)

            # Label the x-ticks with the original indices
            if label is not None:
                plt.xticks(range(len(sorted_x)), sorted_label, rotation=rotation)
            else:
                plt.xticks(range(len(sorted_x)), indices, rotation=rotation)
            
            # Adding the text on top of the bars
            for bar in bars:
                height = bar.get_height()  # Get the height of the bar
                # Place the text at the top of the bar, slightly above
                plt.text(bar.get_x() + bar.get_width() / 2, height + text_offset, f'{height:.{decimal}f}', ha='center', va='bottom')
            
            plt.ylim(0, max(sorted_x) * 1.1)

            if model_rsq:
                # Add text for the sum of values
                plt.text(len(sorted_x) - 1, max(sorted_x), f'Model R²: {round(x_sum, 3)}', 
                    horizontalalignment='right', verticalalignment='top', fontsize=12)
                
            plt.xlabel(xtitle)
            plt.ylabel(ytitle)
            
        else:
            bars = plt.barh(range(len(sorted_x)), sorted_x, color=colors)
            
            # Label the y-ticks with the original indices
            if label is not None:
                plt.yticks(range(len(sorted_x)), sorted_label, rotation=rotation)
            else:
                plt.yticks(range(len(sorted_x)), indices, rotation=rotation)
            plt.gca().invert_yaxis()
            
            # Adding the text beside the bars
            for bar in bars:
                width = bar.get_width()  # Get the width of the bar (since it's horizontal)
                # Place the text to the right of the bar, slightly beyond its end
                plt.text(width + text_offset, bar.get_y() + bar.get_height() / 2, f'{width:.{decimal}f}', 
                        ha='left', va='center')
                
            plt.xlim(0, max(sorted_x) * 1.2)
            
            if model_rsq:
                plt.text(max(sorted_x), len(sorted_x) - 1,f'Model R²: {round(x_sum, 3)}', 
                horizontalalignment='right', verticalalignment='bottom', fontsize=12)
                
            plt.xlabel(ytitle)
            plt.ylabel(xtitle)

        # Add title
        plt.title(title)

        
        if save_name is not None:
            name = save_name + ".pdf"
            plt.savefig(name, bbox_inches='tight')

        plt.show()
        
        plt.close()
    
    # vis.rsq(rsq_res)
    # # Change color
    # vis.rsq(rsq_res, color_map_name="Pastel2")

    # import numpy as np

    # # Generate feature names using list comprehension and format them
    # feature_names = np.array([f"feature{i}" for i in range(1, rsq_res.shape[0]+1)])

    # # Give it a name and rotate
    # vis.rsq(rsq_res, color_map_name="Pastel2", label=feature_names, rotation=45)

    # # Give a horizontal plot, hide model rsq, change the number of features to show
    # vis.rsq(rsq_res, color_map_name="PuBu", horizontal=True, model_rsq=False, max_feature=15, save_name="rsq_eg")
    
    @staticmethod
    def loss(loss, save_ind=None, save_prefix="Shapley loss sample", title="Shapley Loss: Sample", color_map_name="Blues", model_rsq=False, decimal=0, xtitle="Feature Index", ytitle="Loss"):
        """
        Visualize the loss function for each sample
        
        Parameters:
        -loss: multidimensonal loss matrix. usually n*p
        -save_ind: index of sample you want to save, default None.
        -save_prefix: prefix if you want to save
        other parameters inherite from vis_rsq
        """
        def sample_loss(i):
            plt.cla()  # Clear the current axes
            if 0 <= i < loss.shape[0]:  # Check if i is within the valid range
                vis.rsq(loss[i], title=title + " " + str(i), color_map_name=color_map_name, model_rsq=model_rsq, decimal=decimal, xtitle=xtitle, ytitle=ytitle)
        
        # # Slider for quick navigation
        # i_slider = widgets.IntSlider(
        #     value=0,
        #     min=0,
        #     max=x.shape[1]-1,
        #     step=1,
        #     description='Slider:',
        #     continuous_update=False
        # )

        # Text input for precise entry
        i_text = widgets.BoundedIntText(
            value=0,
            min=0,
            max=loss.shape[0]-1,
            step=1,
            description='Sample Index',
            style={'description_width': 'initial'},  # This ensures the description is not cut off
            layout=Layout(width='150px'), 
            continuous_update=True
        )

        # Link the slider and text input to keep them in sync
        # widgets.jslink((i_slider, 'value'), (i_text, 'value'))

        # # Display both the slider and the text input
        # ui = widgets.VBox([i_slider, i_text])
        ui = widgets.VBox([i_text])

        # Use interactive_output to update the plot, linking both controls
        out = widgets.interactive_output(sample_loss, {'i': i_text})

        if save_ind is not None:
            save_name = save_prefix + " " + str(save_ind)
            vis.rsq(loss[save_ind], title=title + " " + str(save_ind), color_map_name=color_map_name, model_rsq=model_rsq, decimal=decimal, xtitle=xtitle, ytitle=ytitle, 
                save_name=save_name)
        else:
            display(ui, out)
        
    # vis.loss(loss)

    # # Find a lovely plot and save it, say for the 5-th sample
    # vis.loss(loss, save_ind=10)
            
    @staticmethod
    def elbow(x, xtitle="Feature Number", ytitle="Explained Variance",
                max_comp=10, title='Explained Variance by Top Features', marker='o', linestyle='--'):
        """
        Construct elbow plot for top features.
        
        Parameters:
        -x: Shapley R squared
        -max_comp: maximum number of components in the plot
        
        Return:
        An elbow plot and the top indices in x that have the highest variance explained
        """
        # Ensure max_comp is not greater than the length of x
        max_comp = int(min(max_comp, len(x)))
        
        # Get indices of the sorted variances in descending order
        indices_sorted_variances = np.argsort(x)[::-1]
        
        # Select the indices corresponding to the highest "max_comp" variances
        selected_indices = indices_sorted_variances[:max_comp]
        
        # Calculate cumulative explained variance for the selected components
        sorted_variances = x[selected_indices]
        #cumulative_explained_variance = np.cumsum(sorted_variances / np.sum(x))
        
        
        plt.plot(range(1, max_comp + 1), sorted_variances, marker=marker, linestyle=linestyle)
        plt.title(title)
        plt.xlabel(xtitle)
        plt.ylabel(ytitle)
        
        # Setting integer ticks on the x-axis
        plt.xticks(range(1, max_comp + 1), range(1, max_comp + 1))
        plt.show()
        
        plt.close()
        
        return selected_indices
    
        # Call the elbow_plot_indices function
         #indices_for_max_comp = vis_elbow(rsq_res, max_comp = 15)
         #indices_for_max_comp
    
    @staticmethod
    def gcorr(x, color_map_name="Blues", horizontal=False, max_feature=10, cutoff=0, title="Generalized Correlation of Features to the Outcome", xtitle="Feature index", ytitle="Generalized Correlation", rotation=0, label=None, decimal=3, save_name=None):
        """
        Visualize Generalized Shapley correlation
        
        Parameters:
        -x: 1 dim array, but we recommend array from shapley rsq OvO ...
        -color_map_name: Color map allows to you take a variety color map from matplotlib to customize your visualization color.
        You can try to use "Pastel1", "Pasterl2", "PuBu", "Reds", "Greens" ...
        -horizontal: horizontal plot or not
        -max_feature: maximum number of features to show
        -cutoff: only value greater than or equal to the cutoff will be displayed
        -xtitle: xtitle, note that it's reversed for horizontal 
        -ytitle: ytitle, note that it's reversed for horizontal 
        -rotation: rotation of the tick
        -label: label for the features
        -title: plot title
        -decimal: decimals to show 
        -save_name: the name of the file if you want to save. None as default without saving.
        """
        vis.rsq(np.sqrt(x), color_map_name=color_map_name, horizontal=horizontal, model_rsq=False, max_feature=max_feature, cutoff=cutoff, title=title, xtitle=xtitle, ytitle=ytitle, rotation=0, label=None, decimal=3, save_name=None)






from qshap import gazer
import sklearn.ensemble
import numpy as np
import time as time
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.preprocessing import Binarizer
import xgboost

from sklearn.datasets import make_regression

# Install the developer's version
# python -m pip install .
# If no priviledge 
# sudo -H python -m pip install .

# Let's start with a toy example with 1000 samples and 1000 features
np.random.seed(0)
x, y, coefficients = make_regression(n_samples=1000, n_features=1000, n_informative=5, coef=True, random_state=0)
#binarizer = Binarizer(threshold=np.median(x))
#x_binary = binarizer.transform(x)

# model fitting
# scikit learn decision tree example, the initial run after import will be slower since numba is comiling
max_depth = 2
tree_regressor = DecisionTreeRegressor(max_depth=max_depth)
tree_fit = tree_regressor.fit(x, y)

start = time.time()
gazer_rsq = gazer(tree_regressor)
rsq_res = gazer.rsq(gazer_rsq, x, y)
end = time.time()
print("time: " + str(end - start))

# Let's check the real R^2
ypred = tree_regressor.predict(x)
sst = np.sum((y - np.mean(y)) ** 2)
sse = np.sum((y - ypred) ** 2)
model_rsq = 1 - sse/sst

print("Treeshap R^2 sum is: " + str(np.sum(rsq_res)))
print("Model R^2 is: " + str(model_rsq) + "\n")
#print("Treeshap R^2 sum and the model R^2 is equal?:  " + str(round(np.sum(rsq_res), 3)==round(model_rsq, 3)) + "!!!!!")

# scikit learn gbdt example. The usage is the same.
# model fitting
max_depth = 4
n_estimators = 50
tree_regressor = GradientBoostingRegressor(n_estimators=n_estimators, max_depth=max_depth)
tree_fit = tree_regressor.fit(x, y)

# start to explain 
start = time.time()
gazer_rsq = gazer(tree_regressor)
rsq_res = gazer.rsq(gazer_rsq, x, y)
end = time.time()
print("time: " + str(end - start))

# Let's check the real R^2
ypred = tree_regressor.predict(x)
sst = np.sum((y - np.mean(y)) ** 2)
sse = np.sum((y - ypred) ** 2)
model_rsq = 1 - sse/sst

print("Treeshap R^2 sum is: " + str(np.sum(rsq_res)))
print("Model R^2 is: " + str(model_rsq) + "\n")

# scikit learn xgboost example. The usage is the same again.
# model fitting
max_depth = 2
n_estimators = 50
tree_regressor = xgboost.XGBRegressor(n_estimators=n_estimators, max_depth=max_depth)
tree_regressor.fit(x, y)


# start to explain 
start = time.time()
gazer_rsq = gazer(tree_regressor)
rsq_res = gazer.rsq(gazer_rsq, x, y)
end = time.time()
print("time: " + str(end - start))

# Let's check the real R^2
ypred = tree_regressor.predict(x)
sst = np.sum((y - np.mean(y)) ** 2)
sse = np.sum((y - ypred) ** 2)
model_rsq = 1 - sse/sst

print("Treeshap R^2 sum is: " + str(np.sum(rsq_res)))
print("Model R^2 is: " + str(model_rsq) + "\n")

np.sum(gazer_rsq.explainer.shap_values(x)*y[:, np.newaxis]/sst, axis=0)

# if you would like to use sampling
start = time.time()
gazer_rsq = gazer(tree_regressor)
rsq_res_frac = gazer.rsq(gazer_rsq, x, y, nfrac=0.5)
end = time.time()
print("time: " + str(end - start))

# Let's check the real R^2
ypred = tree_regressor.predict(x)
sst = np.sum((y - np.mean(y)) ** 2)
sse = np.sum((y - ypred) ** 2)
model_rsq = 1 - sse/sst

print("Treeshap R^2 sum is: " + str(np.sum(rsq_res_frac)))
print("Model R^2 is: " + str(model_rsq) + "\n")


# output generalized correlation (Square root of Shapley R squared)
gcorr_res = gazer.gcorr(rsq_res)

# extract loss example 

# Or simply output both 
rsq_res2 = gazer.rsq(gazer_rsq, x, y, loss_out=True)
rsq_res2.rsq
rsq_res2.loss

# You can extract loss only by, it decompose the loss for each sample
# But directly calling loss doesn't support parallel computing yet
loss_res = gazer.loss(gazer_rsq, x, y)

# Interstingly, you can calculate loss decomposition for arbitrarily sample (which doesn't make sense for R^2)
# This will be particular fast if you only want to several samples among a large dataset
# All you have to do is add the original mean of y
loss_sample1 = gazer.loss(gazer_rsq, x[0:1, :], y[0], y_mean_ori=np.mean(y))
loss_multiple = gazer.loss(gazer_rsq, x[0:99], y[0:99], y_mean_ori=np.mean(y))
# Check if the result match
#print(np.sum(np.abs(loss_res[0] - loss_sample1)))
#print(np.sum(np.abs(loss_res[0:99] - loss_multiple)))

# Visualize rsq 
# First we can import the module
from qshap import vis

# # Generate feature names using list comprehension and format them
feature_names = np.array([f"feature{i}" for i in range(1, rsq_res.shape[0]+1)])

# # Give it a name and rotate
vis.rsq(rsq_res, color_map_name="Pastel2", label=feature_names, rotation=45)

# default
vis.rsq(rsq_res)

# plot up to a certain cutoff
vis.rsq(rsq_res, cutoff=0.1, model_rsq=False)

# change color 
vis.rsq(rsq_res, color_map_name="Pastel2")
# Give a horizontal plot, hide model rsq, change the number of features to show, and you can always save it
vis.rsq(rsq_res, color_map_name="PuBu", horizontal=True, model_rsq=False, max_feature=15, save_name="rsq_eg")

# Elbow plot 
vis.elbow(rsq_res, max_comp=15)

# Vidualize individual loss
# The interactive effect works with jupyter notebook
vis.loss(loss_res)

# Find a lovely plot for one sample and save it, say for the 10-th sample
vis.loss(loss_res, save_ind=10)

# Visualize generalized rsq
vis.gcorr(rsq_res)



from qshap.utils import summarize_tree, simple_tree, tree_summary, weight, store_complex_root, store_complex_v_invc, xgb_formatter, divide_chunks
from qshap.qshap import loss_treeshap

from types import SimpleNamespace
from concurrent.futures import ProcessPoolExecutor
from tqdm import tqdm

import sklearn
import numpy as np
import xgboost
import shap
import warnings
import os
import json
import uuid

class gazer:
    def __init__(self, model):
        self.model = model
        implemented_model = (sklearn.tree.DecisionTreeRegressor,
                                sklearn.ensemble.GradientBoostingRegressor,
                                xgboost.sklearn.XGBRegressor)
        
        if isinstance(model, implemented_model):
            self.explainer = shap.TreeExplainer(model)
            if isinstance(model, (sklearn.tree.DecisionTreeRegressor,sklearn.ensemble.GradientBoostingRegressor)):
                self.max_depth = model.max_depth
            elif isinstance(model, xgboost.sklearn.XGBRegressor):
                # set to default value 6 if max_depth not set by user 
                max_depth_xgb = model.get_params().get("max_depth")
                if max_depth_xgb is not None:
                    self.max_depth = max_depth_xgb
                else:
                    self.max_depth = 6

                unique_filename = str(uuid.uuid4())
                model_filename = f"xgb_model_{unique_filename}.json"
                model.save_model(model_filename)

               # Load the model data
                with open(model_filename, 'r') as file:
                    model_data = json.load(file)
                # Delete it after loading 
                os.remove(model_filename)
                self.base_score = np.float64(model_data["learner"]["learner_model_param"]['base_score'])
                self.xgb_res = xgb_formatter(model_data, self.max_depth)
                del model_data

            # store v_inc * c /d evaluated at complex roots     
            self.store_v_invc = store_complex_v_invc(self.max_depth * 2)
            self.store_z = store_complex_root(self.max_depth * 2)
        else:
            supported_models_str = ', '.join([m.__name__ for m in implemented_model])
            raise NotImplementedError(f"Model not supported yet. Supported models are: {supported_models_str}")
        

    def loss(self, x, y, y_mean_ori=None, progress_bar=True):
        """
        Parameters
        -x: x
        -y: y
        -y_mean_ori: mean of the original
        -progress_bar: whether show the progress bar or not
        """
        max_depth = self.max_depth
        model = self.model
        store_v_invc = self.store_v_invc 
        store_z = self.store_z
        explainer = self.explainer

        if y_mean_ori is None:
            y_mean_ori = np.mean(y)

        if isinstance(model, sklearn.tree.DecisionTreeRegressor):
            summary_tree = summarize_tree(model.tree_)
            loss = loss_treeshap(x, y, summary_tree, store_v_invc, store_z, explainer)

        # GBM 
        elif isinstance(model, sklearn.ensemble.GradientBoostingRegressor):
            ensemble_tree = model.estimators_
            num_tree = len(model)
            staged_predict = list(model.staged_predict(x))
            # learning_rate 
            alpha = model.learning_rate
        
            loss = np.zeros_like(x, dtype=np.float64)
            
            iterator = tqdm(range(num_tree)) if progress_bar else range(num_tree)

            for i in iterator:
                if i==0:
                    res = y - y_mean_ori
                else:
                    res = y - staged_predict[i-1]
                    
                summary_tree = summarize_tree(ensemble_tree[i, 0].tree_)
                explainer = shap.TreeExplainer(ensemble_tree[i, 0])
                loss += loss_treeshap(x, res, summary_tree, store_v_invc, store_z, explainer, alpha)

        # XGBOOST 
        elif isinstance(model, xgboost.sklearn.XGBRegressor):

            xgb_booster = model.get_booster()           
            xgb_res = self.xgb_res
            num_tree = len(xgb_res)

            warnings.filterwarnings("ignore", module="xgb")

            loss = np.zeros_like(x, dtype=np.float64)
            
            iterator = tqdm(range(num_tree)) if progress_bar else range(num_tree)

            for i in iterator:
                # get summary_tree first 
                if i==0:
                    res = y - self.base_score
                else:
                    res = y - model.predict(x, iteration_range=(0, i))
                
                summary_tree = summarize_tree(xgb_res[i])
                explainer = shap.TreeExplainer(xgb_booster[i])
                
                # learning rate is different
                loss += loss_treeshap(x, res, summary_tree, store_v_invc, store_z, explainer, 1)

        return loss
    


    def rsq(self, x, y, loss_out=False, ncore=1, nfrac=None, random_state=42, progress_bar=True):
        """
        Parameters
        -x: the original x
        -y: the original y
        -loss_out: output loss or not
        -nfrac: fraction of samples to sample from, by default use all samples
        -ncore: number of cores to use, with default value 1. It will NOT be beneficial for small datasets and shallow depth.
        -random_state: control random seed for numpy
        -progress_bar: whether show the progress bar or not

        Return
        Shapley R-squared
        """ 
        if nfrac is not None:
            if nfrac <= 0 or nfrac >= 1:
                raise ValueError("Sample fraction (nfrac) must be between (0, 1), use none for no sampling.")
            np.random.seed(random_state)
            sample_size = int(len(x) * nfrac)
            sample_ind = np.random.choice(len(x), sample_size, replace=False)
            x = x[sample_ind]
            y = y[sample_ind]
                  
        max_core = os.cpu_count()
        if ncore == -1:
            ncore = os.cpu_count()
        ncore = min(max_core, ncore)

        explainer = self.explainer
        y_mean_ori = np.mean(y)
        sst = np.sum((y - y_mean_ori) ** 2)
        
        if ncore==1:
            loss = self.loss(x, y, y_mean_ori=y_mean_ori, progress_bar=progress_bar)
        else:
            x_chunks = divide_chunks(x, ncore)
            y_chunks = divide_chunks(y, ncore)

            with ProcessPoolExecutor(max_workers=ncore) as executor:
                # Submit all chunks for processing
                futures = [executor.submit(self.loss, x_chunks[i], y_chunks[i], y_mean_ori, False) for i in range(ncore)]

                # Assign progress bar to a variable
                iterator = tqdm(futures, desc="Processing", total=len(futures)) if progress_bar else futures

                # Wait for all futures to complete and collect results
                results = [future.result() for future in iterator]
            
            loss = np.concatenate(results) 

        rsq = 0 - np.sum(loss, axis=0)/sst

        if loss_out:
            return SimpleNamespace(rsq=rsq, loss=loss)
        else:
            return rsq
        

    def gcorr(rsq_res):
        """
        Parameters
        -rsq_res: the rsq result from calling gazer.rsq

        Return
        Generalized correlation (Square root of Shapley R-squared)
        """
        res = np.sqrt(rsq_res)
        return rsq_res



        


from setuptools import setup, find_packages

setup(
    name='qshap',
    version='0.3.1',
    description='Exact computation of Shapley R-squared in polynomial time',
    long_description=open('README.md').read(),    
    long_description_content_type='text/markdown',  # Specify that the long_description is in Markdown
    author='Zhongli Jiang, Dabao Zhang', 
    author_email='jiang548@purdue.edu, zdb969@hs.uci.edu',
    license="GPL-2.0",
    packages=find_packages(),
    url="https://github.com/catstats/Q-SHAP",
    setup_requires=['numpy'],
    install_requires=['numpy', 'scikit-learn',  'shap', 'numba', 'ipywidgets', 'pandas', 'matplotlib'],
    classifiers=[
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    python_requires='>=3.7', 
    zip_safe=False
)
